;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "speed-type" "20260330.1430"
  "Practice touch and speed typing."
  '((emacs  "27.1")
    (compat "29.1.3"))
  :url "https://github.com/dakra/speed-type"
  :commit "f349c000c4d9541cc0e9dcf072555c7f4c9f9a46"
  :revdesc "f349c000c4d9"
  :keywords '("games")
  :maintainers '(("Daniel Kraus" . "daniel@kraus.my")))
